import keyword
import dataclasses
from collections import defaultdict

import pytest
import zigpy.types as zigpy_t
import zigpy.zdo.types

import zigpy_znp.frames as frames
import zigpy_znp.commands as c
from zigpy_znp import types as t


def test_command_header():
    """Test CommandHeader class."""
    data = b"\x61\x02"
    extra = b"the rest of data\xaa\x55"
    r, rest = t.CommandHeader.deserialize(data + extra)
    assert rest == extra
    assert r.cmd0 == 0x61
    assert r.id == 0x02

    r = t.CommandHeader(0x0261)
    assert r.cmd0 == 0x61
    assert r.id == 0x02

    new1 = r.with_id(0xFF)
    assert new1.id == 0xFF
    assert new1.cmd0 == 0x61

    new2 = r.with_id(0x00)
    assert new2.id == 0x00
    assert new2.cmd0 == 0x61


def test_command_setters():
    """Test setters"""
    # the setter order should not matter
    command = t.CommandHeader(0xFFFF)
    for cmd_type in t.CommandType:
        for subsys in t.Subsystem:
            # There's probably no need to iterate over all 256 possible values
            for cmd_id in (0x00, 0xFF, 0x10, 0x01, 0xF0, 0x0F, 0x22, 0xEE):
                perms = [
                    command.with_id(cmd_id).with_type(cmd_type).with_subsystem(subsys),
                    command.with_type(cmd_type).with_id(cmd_id).with_subsystem(subsys),
                    command.with_type(cmd_type).with_subsystem(subsys).with_id(cmd_id),
                    command.with_subsystem(subsys).with_type(cmd_type).with_id(cmd_id),
                    command.with_subsystem(subsys).with_id(cmd_id).with_type(cmd_type),
                    command.with_id(cmd_id).with_subsystem(subsys).with_type(cmd_type),
                    t.CommandHeader(0xFFFF, id=cmd_id, subsystem=subsys, type=cmd_type),
                ]

                assert len(set(perms)) == 1
                assert perms[0].id == cmd_id
                assert perms[0].subsystem == subsys
                assert perms[0].type == cmd_type


def test_error_code():
    data = b"\x03"
    extra = b"the rest of the owl\x00\xff"

    r, rest = t.ErrorCode.deserialize(data + extra)
    assert rest == extra
    assert r == 0x03
    assert r.name == "INVALID_PARAMETER"

    r, rest = t.ErrorCode.deserialize(b"\xaa" + extra)
    assert rest == extra
    assert r.name == "unknown_0xAA"


def _validate_schema(schema):
    for index, param in enumerate(schema):
        assert isinstance(param.name, str)
        assert param.name.isidentifier()
        assert not keyword.iskeyword(param.name)
        assert isinstance(param.type, type)
        assert isinstance(param.description, str)

        # All optional params must be together at the end
        if param.optional:
            assert all(p.optional for p in schema[index:])

        # Trailing bytes must be at the very end
        if issubclass(param.type, t.TrailingBytes):
            assert not schema[index + 1 :]


def test_commands_schema():
    commands_by_id = defaultdict(list)

    for commands in c.ALL_COMMANDS:
        for cmd in commands:
            if cmd.type == t.CommandType.SREQ:
                assert cmd.type == cmd.Req.header.type
                assert cmd.Rsp.header.type == t.CommandType.SRSP
                assert (
                    cmd.subsystem
                    == cmd.Req.header.subsystem
                    == cmd.Rsp.header.subsystem
                )
                assert isinstance(cmd.Req.header, t.CommandHeader)
                assert isinstance(cmd.Rsp.header, t.CommandHeader)

                assert cmd.Req.Rsp is cmd.Rsp
                assert cmd.Rsp.Req is cmd.Req
                assert cmd.Callback is None

                _validate_schema(cmd.Req.schema)
                _validate_schema(cmd.Rsp.schema)

                commands_by_id[cmd.Req.header].append(cmd.Req)
                commands_by_id[cmd.Rsp.header].append(cmd.Rsp)
            elif cmd.type == t.CommandType.AREQ:
                # we call the AREQ Rsp a Callback
                assert cmd.Rsp is None

                # only one of them can be set
                assert (cmd.Callback is not None) ^ (cmd.Req is not None)

                if cmd.Callback is not None:
                    assert cmd.type == cmd.Callback.header.type
                    assert cmd.subsystem == cmd.Callback.header.subsystem
                    assert isinstance(cmd.Callback.header, t.CommandHeader)

                    _validate_schema(cmd.Callback.schema)

                    commands_by_id[cmd.Callback.header].append(cmd.Callback)
                elif cmd.Req is not None:
                    assert cmd.type == cmd.Req.header.type
                    assert cmd.subsystem == cmd.Req.header.subsystem
                    assert isinstance(cmd.Req.header, t.CommandHeader)

                    _validate_schema(cmd.Req.schema)

                    commands_by_id[cmd.Req.header].append(cmd.Req)
                else:
                    assert False, "Command is empty"  # noqa: B011
            elif cmd.type == t.CommandType.SRSP:
                # The one command like this is RPCError
                assert cmd is c.RPCError.CommandNotRecognized

                assert cmd.type == cmd.Rsp.header.type
                assert cmd.Req is None
                assert cmd.Callback is None
                assert cmd.Rsp.header.type == t.CommandType.SRSP
                assert cmd.subsystem == cmd.Rsp.header.subsystem
                assert isinstance(cmd.Rsp.header, t.CommandHeader)

                _validate_schema(cmd.Rsp.schema)

                commands_by_id[cmd.Rsp.header].append(cmd.Rsp)
            else:
                assert False, "Command has unknown type"  # noqa: B011

    duplicate_commands = {
        cmd: commands for cmd, commands in commands_by_id.items() if len(commands) > 1
    }
    assert not duplicate_commands

    assert len(commands_by_id.keys()) == len(c.COMMANDS_BY_ID.keys())


def test_command_param_binding():
    # No params
    c.SYS.Ping.Req()

    # Invalid param name
    with pytest.raises(KeyError):
        c.SYS.Ping.Rsp(asd=123)

    # Valid param name
    c.SYS.Ping.Rsp(Capabilities=t.MTCapabilities.SYS)

    # Too many params, one valid
    with pytest.raises(KeyError):
        c.SYS.Ping.Rsp(foo="asd", Capabilities=t.MTCapabilities.SYS)

    # Not enough params
    with pytest.raises(KeyError):
        c.SYS.Ping.Rsp()

    # Invalid type
    with pytest.raises(ValueError):
        c.UTIL.TimeAlive.Rsp(Seconds=b"asd")

    # Valid type but invalid value
    with pytest.raises(ValueError):
        c.UTIL.SetPreConfigKey.Req(PreConfigKey=t.KeyData([1, 2, 3]))

    # Coerced numerical type
    a = c.UTIL.TimeAlive.Rsp(Seconds=12)
    b = c.UTIL.TimeAlive.Rsp(Seconds=t.uint32_t(12))

    assert a == b
    assert a.Seconds == b.Seconds
    assert type(a.Seconds) == type(b.Seconds) == t.uint32_t  # noqa: E721

    # Overflowing integer types
    with pytest.raises(ValueError):
        c.UTIL.TimeAlive.Rsp(Seconds=10**20)

    # Integers will not be coerced to enums
    assert t.MTCapabilities.SYS == 0x0001

    with pytest.raises(ValueError):
        c.SYS.Ping.Rsp(Capabilities=0x0001)

    # Parameters can be looked up by name
    ping_rsp = c.SYS.Ping.Rsp(Capabilities=t.MTCapabilities.SYS)
    assert ping_rsp.Capabilities == t.MTCapabilities.SYS

    # Invalid ones cannot
    with pytest.raises(AttributeError):
        ping_rsp.Oops

    # bytes are converted into t.ShortBytes
    cmd = c.SYS.NVWrite.Req(
        SysId=0x12, ItemId=0x3456, SubId=0x7890, Offset=0x0000, Value=b"asdfoo"
    )
    assert isinstance(cmd.Value, t.ShortBytes)

    # Lists are converted to typed LVLists
    c.UTIL.BindAddEntry.Req(
        DstAddrModeAddr=t.AddrModeAddress(mode=t.AddrMode.NWK, address=0x1234),
        DstEndpoint=0x56,
        ClusterIdList=[0x12, 0x45],
    )

    # Type errors within containers are also caught
    with pytest.raises(ValueError):
        c.UTIL.BindAddEntry.Req(
            DstAddrModeAddr=t.AddrModeAddress(mode=t.AddrMode.NWK, address=0x1234),
            DstEndpoint=0x56,
            ClusterIdList=[0x12, 0x457890],  # 0x457890 doesn't fit into a uint8_t
        )


def test_command_optional_params():
    # Optional params values don't need a value
    short_version_rsp = c.SYS.Version.Rsp(
        TransportRev=0,
        ProductId=1,
        MajorRel=2,
        MinorRel=3,
        MaintRel=4,
    )

    # Some can still be passed
    medium_version_rsp = c.SYS.Version.Rsp(
        TransportRev=0, ProductId=1, MajorRel=2, MinorRel=3, MaintRel=4, CodeRevision=5
    )

    # As can all
    long_version_rsp = c.SYS.Version.Rsp(
        TransportRev=0,
        ProductId=1,
        MajorRel=2,
        MinorRel=3,
        MaintRel=4,
        CodeRevision=5,
        BootloaderBuildType=c.sys.BootloaderBuildType.NON_BOOTLOADER_BUILD,
        BootloaderRevision=0xFFFFFFFF,
    )

    short_data = short_version_rsp.to_frame().data
    medium_data = medium_version_rsp.to_frame().data
    long_data = long_version_rsp.to_frame().data

    assert len(long_data) == len(medium_data) + 5 == len(short_data) + 9

    assert long_data.startswith(medium_data)
    assert medium_data.startswith(short_data)

    # Deserialization is greedy
    Version = c.SYS.Version.Rsp
    assert Version.from_frame(long_version_rsp.to_frame()) == long_version_rsp
    assert Version.from_frame(medium_version_rsp.to_frame()) == medium_version_rsp
    assert Version.from_frame(short_version_rsp.to_frame()) == short_version_rsp

    # Deserialization still fails if the frame is incomplete
    with pytest.raises(ValueError):
        Version.from_frame(
            frames.GeneralFrame(
                header=long_version_rsp.to_frame().header, data=long_data[:-1]
            )
        )

    # Deserialization will fail if the frame is incomplete but has no truncated fields
    with pytest.raises(ValueError):
        Version.from_frame(
            frames.GeneralFrame(
                header=long_version_rsp.to_frame().header, data=long_data[:4]
            )
        )

    with pytest.raises(ValueError):
        Version.from_frame(
            frames.GeneralFrame(
                header=long_version_rsp.to_frame().header, data=long_data + b"\x00"
            )
        )


def test_command_optional_params_failures():
    with pytest.raises(KeyError):
        # Optional params cannot be skipped over
        c.SYS.Version.Rsp(
            TransportRev=0,
            ProductId=1,
            MajorRel=2,
            MinorRel=3,
            MaintRel=4,
            # CodeRevision=5,
            BootloaderBuildType=c.sys.BootloaderBuildType.BUILT_AS_HEX,
        )

    # Unless it's a partial command
    partial = c.SYS.Version.Rsp(
        TransportRev=0,
        ProductId=1,
        MajorRel=2,
        MinorRel=3,
        MaintRel=4,
        # CodeRevision=5,
        BootloaderBuildType=c.sys.BootloaderBuildType.BUILT_AS_HEX,
        partial=True,
    )

    # In which case, it cannot be serialized
    with pytest.raises(ValueError):
        partial.to_frame()


def test_simple_descriptor():
    lvlist16_type = zigpy_t.LVList[t.uint16_t]

    simple_descriptor = zigpy.zdo.types.SimpleDescriptor()
    simple_descriptor.endpoint = zigpy_t.uint8_t(1)
    simple_descriptor.profile = zigpy_t.uint16_t(260)
    simple_descriptor.device_type = zigpy_t.uint16_t(257)
    simple_descriptor.device_version = zigpy_t.uint8_t(0)
    simple_descriptor.input_clusters = lvlist16_type([0, 3, 4, 5, 6, 8, 2821, 1794])
    simple_descriptor.output_clusters = lvlist16_type([10, 25])

    c1 = c.ZDO.SimpleDescRsp.Callback(
        Src=t.NWK(0x1234),
        Status=t.ZDOStatus.SUCCESS,
        NWK=t.NWK(0x1234),
        SimpleDescriptor=simple_descriptor,
    )

    sp_simple_descriptor = zigpy.zdo.types.SizePrefixedSimpleDescriptor()
    sp_simple_descriptor.endpoint = zigpy_t.uint8_t(1)
    sp_simple_descriptor.profile = zigpy_t.uint16_t(260)
    sp_simple_descriptor.device_type = zigpy_t.uint16_t(257)
    sp_simple_descriptor.device_version = zigpy_t.uint8_t(0)
    sp_simple_descriptor.input_clusters = lvlist16_type([0, 3, 4, 5, 6, 8, 2821, 1794])
    sp_simple_descriptor.output_clusters = lvlist16_type([10, 25])

    c2 = c.ZDO.SimpleDescRsp.Callback(
        Src=t.NWK(0x1234),
        Status=t.ZDOStatus.SUCCESS,
        NWK=t.NWK(0x1234),
        SimpleDescriptor=sp_simple_descriptor,
    )

    assert c1.to_frame() == c2.to_frame()
    # assert c1 == c2


def test_command_str_repr():
    command = c.UTIL.BindAddEntry.Req(
        DstAddrModeAddr=t.AddrModeAddress(mode=t.AddrMode.NWK, address=0x1234),
        DstEndpoint=0x56,
        ClusterIdList=[0x12, 0x34],
    )

    assert repr(command) == str(command)
    assert str([command]) == f"[{str(command)}]"


def test_command_immutability():
    command1 = c.SYS.NVWrite.Req(
        partial=True, SysId=None, ItemId=0x1234, SubId=None, Offset=None, Value=None
    )

    command2 = c.SYS.NVWrite.Req(
        partial=True, SysId=None, ItemId=0x1234, SubId=None, Offset=None, Value=None
    )

    d = {command1: True}

    assert command1 == command2
    assert command2 in d
    assert {command1: True} == {command2: True}

    with pytest.raises(RuntimeError):
        command1.partial = False

    with pytest.raises(RuntimeError):
        command1.SysId = 0x10

    with pytest.raises(RuntimeError):
        command1.ItemId = 0x1234

    with pytest.raises(RuntimeError):
        del command1.ItemId

    assert command1 == command2


def test_command_serialization():
    command = c.SYS.NVWrite.Req(
        SysId=0x12, ItemId=0x3456, SubId=0x7890, Offset=0x0000, Value=b"asdfoo"
    )
    frame = command.to_frame()

    assert frame.data == bytes.fromhex("12 5634 9078 0000 06") + b"asdfoo"

    # Partial frames cannot be serialized
    with pytest.raises(ValueError):
        partial1 = c.SYS.NVWrite.Req(partial=True, SysId=0x12)
        partial1.to_frame()

    # Partial frames cannot be serialized, even if all params are filled out
    with pytest.raises(ValueError):
        partial2 = c.SYS.NVWrite.Req(
            partial=True, SysId=None, ItemId=0x1234, SubId=None, Offset=None, Value=None
        )
        partial2.to_frame()


def test_command_equality():
    command1 = c.SYS.NVWrite.Req(
        SysId=0x12, ItemId=0x3456, SubId=0x7890, Offset=0x00, Value=b"asdfoo"
    )

    command2 = c.SYS.NVWrite.Req(
        SysId=0x12, ItemId=0x3456, SubId=0x7890, Offset=0x00, Value=b"asdfoo"
    )

    command3 = c.SYS.NVWrite.Req(
        SysId=0xFF, ItemId=0x3456, SubId=0x7890, Offset=0x00, Value=b"asdfoo"
    )

    assert command1 == command1
    assert command1.matches(command1)
    assert command2 == command1
    assert command1 == command2

    assert command1 != command3
    assert command3 != command1

    assert command1.matches(command2)  # Matching is a superset of equality
    assert command2.matches(command1)
    assert not command1.matches(command3)
    assert not command3.matches(command1)

    assert not command1.matches(c.SYS.NVWrite.Req(partial=True))
    assert c.SYS.NVWrite.Req(partial=True).matches(command1)

    # parameters can be specified explicitly as None
    assert c.SYS.NVWrite.Req(partial=True, SubId=None).matches(command1)
    assert c.SYS.NVWrite.Req(partial=True, SubId=0x7890).matches(command1)
    assert not c.SYS.NVWrite.Req(partial=True, SubId=123).matches(command1)

    # Different frame types do not match, even if they have the same structure
    assert not c.SYS.NVWrite.Rsp(Status=t.Status.SUCCESS).matches(
        c.SYS.NVDelete.Rsp(partial=True)
    )


def test_command_deserialization():
    command = c.SYS.NVWrite.Req(
        SysId=0x12, ItemId=0x3456, SubId=0x7890, Offset=0x00, Value=b"asdfoo"
    )

    assert type(command).from_frame(command.to_frame()) == command
    assert command.to_frame() == type(command).from_frame(command.to_frame()).to_frame()

    # Deserialization fails if there is unparsed data at the end of the frame
    frame = command.to_frame()
    bad_frame = dataclasses.replace(frame, data=frame.data + b"\x00")

    with pytest.raises(ValueError):
        type(command).from_frame(bad_frame)

    # Deserialization fails if you attempt to deserialize the wrong frame
    with pytest.raises(ValueError):
        c.SYS.NVWrite.Req.from_frame(c.SYS.Ping.Req().to_frame())


def test_command_not_recognized():
    command = c.RPCError.CommandNotRecognized.Rsp(
        ErrorCode=c.rpc_error.ErrorCode.InvalidSubsystem,
        RequestHeader=t.CommandHeader(0xABCD),
    )

    transport_frame = frames.TransportFrame(command.to_frame())

    assert transport_frame.serialize()[:-1] == bytes.fromhex("FE  03  60 00  01  CD AB")


def test_command_replace_normal():
    command1 = c.SYS.NVWrite.Req(
        SysId=0x12, ItemId=0x3456, SubId=0x7890, Offset=0x00, Value=b"asdfoo"
    )

    command2 = c.SYS.NVWrite.Req(
        SysId=0x13, ItemId=0x3456, SubId=0x7890, Offset=0x00, Value=b"asdfoos"
    )

    assert command1.replace() == command1
    assert command1.replace(SysId=0x13, Value=b"asdfoos") == command2


def test_command_replace_partial():
    command1 = c.SYS.NVWrite.Req(partial=True, SysId=0x12)

    command2 = c.SYS.NVWrite.Req(partial=True, SysId=0x13)

    assert command1.replace() == command1
    assert command1.replace(SysId=0x13) == command2


def test_command_possibly_empty_payload():
    class TestSubsystem(t.CommandsBase, subsystem=t.Subsystem.SYS):
        Test = t.CommandDef(
            t.CommandType.AREQ,
            0x00,
            rsp_schema=(t.Param("Data", t.Bytes, "Can be any length"),),
        )

    Test = TestSubsystem.Test.Callback

    assert Test.from_frame(
        frames.GeneralFrame(header=Test.header, data=b"test")
    ) == Test(Data=t.Bytes(b"test"))

    assert Test.from_frame(frames.GeneralFrame(header=Test.header, data=b"")) == Test(
        Data=t.Bytes(b"")
    )



from __future__ import annotations

import logging

import zigpy.zdo
import zigpy.device
import zigpy.application

LOGGER = logging.getLogger(__name__)

NWK_UPDATE_LOOP_DELAY = 1


class ZNPCoordinator(zigpy.device.Device):
    """
    Coordinator zigpy device that keeps track of our endpoints and clusters.
    """

    @property
    def manufacturer(self):
        return "Texas Instruments"

    @property
    def model(self):
        if self.application._znp.version > 3.0:
            model = "CC1352/CC2652"
            version = "3.30+"
        else:
            model = "CC2538" if self.application._znp.nvram.align_structs else "CC2531"
            version = "Home 1.2" if self.application._znp.version == 1.2 else "3.0.x"

        return f"{model}, Z-Stack {version} (build {self.application._zstack_build_id})"

    def request(
        self,
        profile,
        cluster,
        src_ep,
        dst_ep,
        sequence,
        data,
        expect_reply=True,
        # Extend the default timeout
        timeout=2 * zigpy.device.APS_REPLY_TIMEOUT,
        use_ieee=False,
    ):
        """
        Normal `zigpy.device.Device:request` except its default timeout is longer.
        """

        return super().request(
            profile,
            cluster,
            src_ep,
            dst_ep,
            sequence,
            data,
            expect_reply=expect_reply,
            timeout=timeout,
            use_ieee=use_ieee,
        )



import ast

import pytest

import zigpy_znp.types as t


def test_serialize_list():
    class TestList(t.LVList, item_type=t.uint8_t, length_type=t.uint8_t):
        pass

    assert (
        t.serialize_list([t.uint8_t(0xF0), t.Bytes(b"asd"), TestList([0xAB, 0xCD])])
        == b"\xF0asd\x02\xAB\xCD"
    )

    assert t.serialize_list([]) == b""


def test_enum_uint():
    class TestEnum(t.enum_flag_uint16):
        ALL = 0xFFFF
        CH_1 = 0x0001
        CH_2 = 0x0002
        CH_3 = 0x0004
        CH_5 = 0x0008
        CH_6 = 0x0010
        CH_Z = 0x8000

    extra = b"The rest of the data\x55\xaa"
    data = b"\x12\x80"

    r, rest = TestEnum.deserialize(data + extra)
    assert rest == extra
    assert r == 0x8012
    assert r == (TestEnum.CH_2 | TestEnum.CH_6 | TestEnum.CH_Z)

    assert r.serialize() == data
    assert TestEnum(0x8012).serialize() == data


def test_abstract_ints():
    assert issubclass(t.uint8_t, t.uint_t)
    assert not issubclass(t.uint8_t, t.int_t)
    assert t.int_t._signed is True
    assert t.uint_t._signed is False

    with pytest.raises(TypeError):
        t.int_t(0)

    with pytest.raises(TypeError):
        t.FixedIntType(0)


def test_int_too_short():
    with pytest.raises(ValueError):
        t.uint8_t.deserialize(b"")

    with pytest.raises(ValueError):
        t.uint16_t.deserialize(b"\x00")


def test_int_out_of_bounds():
    with pytest.raises(ValueError):
        t.uint8_t(-1)

    with pytest.raises(ValueError):
        t.uint8_t(0xFF + 1)


def test_bytes():
    data = b"abcde\x00\xff"

    r, rest = t.Bytes.deserialize(data)
    assert rest == b""
    assert r == data

    assert r.serialize() == data

    assert str(r) == repr(r) == "b'\\x61\\x62\\x63\\x64\\x65\\x00\\xFF'"

    # Ensure we don't make any mistakes formatting the bytes
    all_bytes = t.Bytes(bytes(range(0, 255 + 1)))
    long_repr = repr(all_bytes)
    assert ast.literal_eval(long_repr) == ast.literal_eval(bytes.__repr__(all_bytes))
    assert all_bytes == ast.literal_eval(long_repr)


def test_longbytes():
    data = b"abcde\x00\xff" * 50
    extra = b"\xffrest of the data\x00"

    r, rest = t.LongBytes.deserialize(len(data).to_bytes(2, "little") + data + extra)
    assert rest == extra
    assert r == data

    assert r.serialize() == len(data).to_bytes(2, "little") + data

    with pytest.raises(ValueError):
        t.LongBytes.deserialize(b"\x01")

    with pytest.raises(ValueError):
        t.LongBytes.deserialize(b"\x01\x00")

    with pytest.raises(ValueError):
        t.LongBytes.deserialize(len(data).to_bytes(2, "little") + data[:-1])


def test_lvlist():
    class TestList(t.LVList, item_type=t.uint8_t, length_type=t.uint8_t):
        pass

    d, r = TestList.deserialize(b"\x0412345")
    assert r == b"5"
    assert d == list(map(ord, "1234"))
    assert TestList.serialize(d) == b"\x041234"

    assert isinstance(d, TestList)

    with pytest.raises(ValueError):
        TestList([1, 2, 0xFFFF, 4]).serialize()


def test_lvlist_too_short():
    class TestList(t.LVList, item_type=t.uint8_t, length_type=t.uint8_t):
        pass

    with pytest.raises(ValueError):
        TestList.deserialize(b"")

    with pytest.raises(ValueError):
        TestList.deserialize(b"\x04123")


def test_hex_repr():
    class NwkAsHex(t.uint16_t, hex_repr=True):
        pass

    nwk = NwkAsHex(0x123A)
    assert str(nwk) == "0x123A"
    assert repr(nwk) == "0x123A"

    assert str([nwk]) == "[0x123A]"
    assert repr([nwk]) == "[0x123A]"

    # You can turn it off as well
    class NwkWithoutHex(NwkAsHex, hex_repr=False):
        pass

    nwk = NwkWithoutHex(1234)
    assert str(nwk) == "1234"
    assert repr(nwk) == "1234"

    assert str([nwk]) == "[1234]"
    assert repr([nwk]) == "[1234]"


def test_fixed_list():
    class TestList(t.FixedList, item_type=t.uint16_t, length=3):
        pass

    with pytest.raises(ValueError):
        r = TestList([1, 2, 3, 0x55AA])
        r.serialize()

    with pytest.raises(ValueError):
        r = TestList([1, 2])
        r.serialize()

    r = TestList([1, 2, 3])

    assert r.serialize() == b"\x01\x00\x02\x00\x03\x00"


def test_fixed_list_deserialize():
    class TestList(t.FixedList, length=3, item_type=t.uint16_t):
        pass

    data = b"\x34\x12\x55\xaa\x89\xab"
    extra = b"\x00\xff"

    r, rest = TestList.deserialize(data + extra)
    assert rest == extra
    assert r[0] == 0x1234
    assert r[1] == 0xAA55
    assert r[2] == 0xAB89


def test_enum_instance_types():
    class TestEnum(t.enum_uint8):
        Member = 0x00

    assert TestEnum._member_type_ is t.uint8_t
    assert type(TestEnum.Member.value) is t.uint8_t



import enum

import pytest
from voluptuous import Invalid

import zigpy_znp.config as conf


def test_EnumValue():
    class TestEnum(enum.Enum):
        foo = 123
        BAR = 456

    assert conf.EnumValue(TestEnum)("foo") == TestEnum.foo
    assert conf.EnumValue(TestEnum)("BAR") == TestEnum.BAR

    assert conf.EnumValue(TestEnum, transformer=str.lower)("FOO") == TestEnum.foo
    assert conf.EnumValue(TestEnum, transformer=str.upper)("bar") == TestEnum.BAR

    assert conf.EnumValue(TestEnum)(TestEnum.foo) == TestEnum.foo
    assert conf.EnumValue(TestEnum)(TestEnum.BAR) == TestEnum.BAR


def test_pin_states_same_lengths():
    # Bare schema works
    conf.CONFIG_SCHEMA(
        {
            conf.CONF_DEVICE: {conf.CONF_DEVICE_PATH: "/dev/null"},
            conf.CONF_ZNP_CONFIG: {},
        }
    )

    # So does one with explicitly specified pin states
    config = conf.CONFIG_SCHEMA(
        {
            conf.CONF_DEVICE: {conf.CONF_DEVICE_PATH: "/dev/null"},
            conf.CONF_ZNP_CONFIG: {
                conf.CONF_CONNECT_RTS_STATES: ["on", True, 0, 0, 0, 1, 1],
                conf.CONF_CONNECT_DTR_STATES: ["off", False, 1, 0, 0, 1, 1],
            },
        }
    )

    assert config[conf.CONF_ZNP_CONFIG][conf.CONF_CONNECT_RTS_STATES] == [
        True,
        True,
        False,
        False,
        False,
        True,
        True,
    ]
    assert config[conf.CONF_ZNP_CONFIG][conf.CONF_CONNECT_DTR_STATES] == [
        False,
        False,
        True,
        False,
        False,
        True,
        True,
    ]


def test_pin_states_different_lengths():
    # They must be the same length
    with pytest.raises(Invalid):
        conf.CONFIG_SCHEMA(
            {
                conf.CONF_DEVICE: {conf.CONF_DEVICE_PATH: "/dev/null"},
                conf.CONF_ZNP_CONFIG: {
                    conf.CONF_CONNECT_RTS_STATES: [1, 1, 0],
                    conf.CONF_CONNECT_DTR_STATES: [1, 1],
                },
            }
        )



import pytest

import zigpy_znp.types as t
import zigpy_znp.commands as c


def test_status():
    """Test Status enum."""

    assert t.Status.SUCCESS == 0
    assert t.Status.FAILURE == 1

    extra = b"the rest of the owl\xaa\x55\x00\xff"

    r, rest = t.Status.deserialize(b"\x00" + extra)
    assert rest == extra
    assert r == 0
    assert r == t.Status.SUCCESS
    assert r.name == "SUCCESS"

    r, rest = t.Status.deserialize(b"\x33" + extra)
    assert rest == extra
    assert r == 0x33
    assert r.value == 0x33
    assert r.name == "unknown_0x33"


def test_addr_mode_address():
    """Test Addr mode address."""

    data = b"\x03\x00\x01\x02\x03\x04\x05\x06\x07"
    extra = b"The rest of the data\x55\xaa"

    # IEEE
    r, rest = t.AddrModeAddress.deserialize(data + extra)
    assert rest == extra
    assert r.mode == t.AddrMode.IEEE
    assert r.address == t.EUI64(range(8))
    assert r.serialize() == data
    assert t.AddrModeAddress(r) == r
    assert r != object()

    # NWK
    data = b"\x02\xaa\x55\x02\x03\x04\x05\x06\x07"
    r, rest = t.AddrModeAddress.deserialize(data + extra)
    assert rest == extra
    assert r.mode == t.AddrMode.NWK
    assert r.address == t.NWK(0x55AA)
    assert r.serialize()[:3] == data[:3]
    assert len(r.serialize()) == 9
    assert t.AddrModeAddress(r) == r

    # Group
    data = b"\x01\xcd\xab\x02\x03\x04\x05\x06\x07"
    r, rest = t.AddrModeAddress.deserialize(data + extra)
    assert rest == extra
    assert r.mode == t.AddrMode.Group
    assert r.address == t.NWK(0xABCD)
    assert r.serialize()[:3] == data[:3]
    assert len(r.serialize()) == 9
    assert t.AddrModeAddress(r) == r

    # Broadcast
    data = b"\x0f\xfe\xff\x02\x03\x04\x05\x06\x07"
    r, rest = t.AddrModeAddress.deserialize(data + extra)
    assert rest == extra
    assert r.mode == t.AddrMode.Broadcast
    assert r.address == t.NWK(0xFFFE)
    assert r.serialize()[:3] == data[:3]
    assert len(r.serialize()) == 9
    assert t.AddrModeAddress(r) == r

    with pytest.raises(ValueError):
        # 0xab is not a valid mode
        data = b"\xAB\xaa\x55\x02\x03\x04\x05\x06\x07"
        t.AddrModeAddress.deserialize(data)

    with pytest.raises(ValueError):
        # NOT_PRESENT is a valid AddrMode member but it is not a valid AddrModeAddress
        data = b"\x00\xaa\x55\x02\x03\x04\x05\x06\x07"
        t.AddrModeAddress.deserialize(data)

    # Bytes at the end for NWK address mode are ignored
    data1 = b"\x02\x0E\xAD" + b"\xC0\x8C\x97\x83\xB0\x20\x33"
    data2 = b"\x02\x0E\xAD" + b"\x3F\xB9\x5B\x64\x20\x86\xD6"

    r1, _ = t.AddrModeAddress.deserialize(data1)
    r2, _ = t.AddrModeAddress.deserialize(data2)

    assert r1 == r2

    # All of the bytes are used for IEEE address mode
    data1 = b"\x02\x0E\xAD\xC0\x8C\x97\x83\xB0\x20\x33"
    data2 = b"\x02\x0E\xAD\x3F\xB9\x5B\x64\x20\x86\xD6"

    r3, _ = t.AddrModeAddress.deserialize(b"\x03" + data1[1:])
    r4, _ = t.AddrModeAddress.deserialize(b"\x03" + data2[1:])

    assert r3 != r4


def test_missing_status_enum():
    class TestEnum(t.MissingEnumMixin, t.enum_uint8):
        Member = 0x00

    assert 0xFF not in list(TestEnum)
    assert isinstance(TestEnum(0xFF), TestEnum)
    assert TestEnum(0xFF).value == 0xFF
    assert type(TestEnum(0xFF).value) is t.uint8_t

    # Missing members that don't fit can't be created
    with pytest.raises(ValueError):
        TestEnum(0xFF + 1)

    # Missing members that aren't integers can't be created
    with pytest.raises(ValueError):
        TestEnum("0xFF")


def test_zdo_nullable_node_descriptor():
    desc1, data = c.zdo.NullableNodeDescriptor.deserialize(b"\x00")

    # Old-style zigpy structs
    assert all(value is None for field, value in desc1.assigned_fields())

    assert not data
    assert desc1.serialize() == b"\x00"

    desc2 = c.zdo.NullableNodeDescriptor(1, 2, 3, 4, 5, 6, 7, 8, 9)
    desc3, data = c.zdo.NullableNodeDescriptor.deserialize(desc2.serialize())

    assert not data
    assert desc2.serialize() == desc3.serialize()


def test_missing_enum_mixin():
    class TestEnum(t.MissingEnumMixin, t.enum_uint8):
        FOO = 0x01

    assert TestEnum(0x01) == 0x01 == TestEnum.FOO
    assert TestEnum(0x02) == 0x02
    assert 0x02 not in TestEnum._value2member_map_



from zigpy_znp.types import nvids


def _test_nvid_uniqueness(nvids_table):
    seen = set()

    for attr in nvids_table:
        assert attr.value not in seen
        seen.add(attr.value)

    assert len(seen) == len(nvids_table)


def test_nvid_uniqueness():
    _test_nvid_uniqueness(nvids.ZclPortNvIds)
    _test_nvid_uniqueness(nvids.ExNvIds)
    _test_nvid_uniqueness(nvids.OsalNvIds)



import typing

import pytest

import zigpy_znp.types as t

if typing.TYPE_CHECKING:
    import typing_extensions


def test_struct_fields():
    class TestStruct(t.CStruct):
        a: t.uint8_t
        b: t.uint16_t

    assert len(TestStruct.fields) == 2

    assert TestStruct.fields.a.name == "a"
    assert TestStruct.fields.a.type == t.uint8_t

    assert TestStruct.fields.b.name == "b"
    assert TestStruct.fields.b.type == t.uint16_t


def test_struct_field_values():
    class TestStruct(t.CStruct):
        a: t.uint8_t
        b: t.uint16_t

    struct = TestStruct(a=1, b=2)
    assert struct.a == 1
    assert isinstance(struct.a, t.uint8_t)

    assert struct.b == 2
    assert isinstance(struct.b, t.uint16_t)

    # Invalid values can't be passed during construction
    with pytest.raises(ValueError):
        TestStruct(a=1, b=2**32)

    struct2 = TestStruct()
    struct2.a = 1
    struct2.b = 2

    assert struct == struct2
    assert struct.serialize() == struct2.serialize()


def test_struct_methods_and_constants():
    class TestStruct(t.CStruct):
        a: t.uint8_t
        b: t.uint16_t

        def method(self):
            return self.a + self.b

        def annotated_method(self: "TestStruct") -> int:
            return self.method()

        CONSTANT1 = 1
        constant2 = "foo"
        _constant3 = "bar"

    assert len(TestStruct.fields) == 2
    assert TestStruct.fields.a == t.CStructField(name="a", type=t.uint8_t)
    assert TestStruct.fields.b == t.CStructField(name="b", type=t.uint16_t)

    assert TestStruct.CONSTANT1 == 1
    assert TestStruct.constant2 == "foo"
    assert TestStruct._constant3 == "bar"

    assert TestStruct(a=1, b=2).method() == 3


def test_struct_nesting():
    class Outer(t.CStruct):
        e: t.uint32_t

    class TestStruct(t.CStruct):
        class Inner(t.CStruct):
            c: t.uint16_t

        a: t.uint8_t
        b: Inner
        d: Outer

    assert len(TestStruct.fields) == 3
    assert TestStruct.fields.a == t.CStructField(name="a", type=t.uint8_t)
    assert TestStruct.fields.b == t.CStructField(name="b", type=TestStruct.Inner)
    assert TestStruct.fields.d == t.CStructField(name="d", type=Outer)

    assert len(TestStruct.Inner.fields) == 1
    assert TestStruct.Inner.fields.c == t.CStructField(name="c", type=t.uint16_t)

    struct = TestStruct(a=1, b=TestStruct.Inner(c=2), d=Outer(e=3))
    assert struct.a == 1
    assert struct.b.c == 2
    assert struct.d.e == 3


def test_struct_aligned_serialization_deserialization():
    class TestStruct(t.CStruct):
        a: t.uint8_t
        # One padding byte here
        b: t.uint16_t
        # No padding here
        c: t.uint32_t  # largest type, so the struct is 32 bit aligned
        d: t.uint8_t
        # Three padding bytes here
        e: t.uint32_t
        f: t.uint8_t
        # Three more to make the struct 32 bit aligned

    assert TestStruct.get_alignment(align=False) == 1
    assert TestStruct.get_alignment(align=True) == 32 // 8
    assert TestStruct.get_size(align=False) == (1 + 2 + 4 + 1 + 4 + 1)
    assert TestStruct.get_size(align=True) == (1 + 2 + 4 + 1 + 4 + 1) + (1 + 3 + 3)

    expected = b""
    expected += t.uint8_t(1).serialize()
    expected += b"\xFF" + t.uint16_t(2).serialize()
    expected += t.uint32_t(3).serialize()
    expected += t.uint8_t(4).serialize()
    expected += b"\xFF\xFF\xFF" + t.uint32_t(5).serialize()
    expected += t.uint8_t(6).serialize()
    expected += b"\xFF\xFF\xFF"

    struct = TestStruct(a=1, b=2, c=3, d=4, e=5, f=6)
    assert struct.serialize(align=True) == expected

    struct2, remaining = TestStruct.deserialize(expected + b"test", align=True)
    assert remaining == b"test"
    assert struct == struct2

    with pytest.raises(ValueError):
        TestStruct.deserialize(expected[:-1], align=True)


def test_struct_aligned_nested_serialization_deserialization():
    class Inner(t.CStruct):
        _padding_byte = b"\xCD"

        c: t.uint8_t
        d: t.uint32_t
        e: t.uint8_t

    class TestStruct(t.CStruct):
        _padding_byte = b"\xAB"

        a: t.uint8_t
        b: Inner
        f: t.uint16_t

    expected = b""
    expected += t.uint8_t(1).serialize()

    # Inner struct
    expected += b"\xAB\xAB\xAB" + t.uint8_t(2).serialize()
    expected += b"\xCD\xCD\xCD" + t.uint32_t(3).serialize()
    expected += t.uint8_t(4).serialize()
    expected += b"\xCD\xCD\xCD"  # Aligned to 4 bytes

    expected += t.uint16_t(5).serialize()
    expected += b"\xAB\xAB"  # Also aligned to 4 bytes due to inner struct

    struct = TestStruct(a=1, b=Inner(c=2, d=3, e=4), f=5)
    assert struct.serialize(align=True) == expected

    struct2, remaining = TestStruct.deserialize(expected + b"test", align=True)
    assert remaining == b"test"
    assert struct == struct2


def test_struct_unaligned_serialization_deserialization():
    class TestStruct(t.CStruct):
        a: t.uint8_t
        b: t.uint16_t
        c: t.uint32_t
        d: t.uint8_t
        e: t.uint32_t
        f: t.uint8_t

    expected = b""
    expected += t.uint8_t(1).serialize()
    expected += t.uint16_t(2).serialize()
    expected += t.uint32_t(3).serialize()
    expected += t.uint8_t(4).serialize()
    expected += t.uint32_t(5).serialize()
    expected += t.uint8_t(6).serialize()

    struct = TestStruct(a=1, b=2, c=3, d=4, e=5, f=6)

    assert struct.serialize(align=False) == expected

    struct2, remaining = TestStruct.deserialize(expected + b"test", align=False)
    assert remaining == b"test"
    assert struct == struct2

    with pytest.raises(ValueError):
        TestStruct.deserialize(expected[:-1], align=False)


def test_struct_equality():
    class InnerStruct(t.CStruct):
        c: t.EUI64

    class TestStruct(t.CStruct):
        a: t.uint8_t
        b: InnerStruct

    class TestStruct2(t.CStruct):
        a: t.uint8_t
        b: InnerStruct

    s1 = TestStruct(a=2, b=InnerStruct(c=t.EUI64.convert("00:00:00:00:00:00:00:00")))
    s2 = TestStruct(a=2, b=InnerStruct(c=t.EUI64.convert("00:00:00:00:00:00:00:00")))
    s3 = TestStruct2(a=2, b=InnerStruct(c=t.EUI64.convert("00:00:00:00:00:00:00:00")))

    assert s1 == s2
    assert s1.replace(a=3) != s1
    assert s1.replace(a=3).replace(a=2) == s1

    assert s1 != s3
    assert s1.serialize() == s3.serialize()

    assert TestStruct(s1) == s1
    assert TestStruct(a=s1.a, b=s1.b) == s1

    with pytest.raises(ValueError):
        TestStruct(s1, b=InnerStruct(s1.b))

    with pytest.raises(ValueError):
        TestStruct2(s1)


def test_struct_repr():
    class TestStruct(t.CStruct):
        a: t.uint8_t
        b: t.uint32_t

    assert str(TestStruct(a=1, b=2)) == "TestStruct(a=1, b=2)"
    assert str([TestStruct(a=1, b=2)]) == "[TestStruct(a=1, b=2)]"


def test_struct_bad_fields():
    with pytest.raises(TypeError):

        class TestStruct(t.CStruct):
            a: t.uint8_t
            b: int


def test_struct_incomplete_serialization():
    class TestStruct(t.CStruct):
        a: t.uint8_t
        b: t.uint8_t

    TestStruct(a=1, b=2).serialize()

    with pytest.raises(ValueError):
        TestStruct(a=1, b=None).serialize()

    with pytest.raises(ValueError):
        TestStruct(a=1).serialize()

    struct = TestStruct(a=1, b=2)
    struct.b = object()

    with pytest.raises(ValueError):
        struct.serialize()


def test_old_nib_deserialize():
    PaddingByte: typing_extensions.TypeAlias = t.uint8_t

    class NwkState16(t.enum_uint16):
        NWK_INIT = 0
        NWK_JOINING_ORPHAN = 1
        NWK_DISC = 2
        NWK_JOINING = 3
        NWK_ENDDEVICE = 4
        PAN_CHNL_SELECTION = 5
        PAN_CHNL_VERIFY = 6
        PAN_STARTING = 7
        NWK_ROUTER = 8
        NWK_REJOINING = 9

    class OldNIB(t.CStruct):
        SequenceNum: t.uint8_t
        PassiveAckTimeout: t.uint8_t
        MaxBroadcastRetries: t.uint8_t
        MaxChildren: t.uint8_t
        MaxDepth: t.uint8_t
        MaxRouters: t.uint8_t
        dummyNeighborTable: t.uint8_t
        BroadcastDeliveryTime: t.uint8_t
        ReportConstantCost: t.uint8_t
        RouteDiscRetries: t.uint8_t
        dummyRoutingTable: t.uint8_t
        SecureAllFrames: t.uint8_t
        SecurityLevel: t.uint8_t
        SymLink: t.uint8_t
        CapabilityFlags: t.uint8_t
        PaddingByte0: PaddingByte
        TransactionPersistenceTime: t.uint16_t
        nwkProtocolVersion: t.uint8_t
        RouteDiscoveryTime: t.uint8_t
        RouteExpiryTime: t.uint8_t
        PaddingByte1: PaddingByte
        nwkDevAddress: t.NWK
        nwkLogicalChannel: t.uint8_t
        PaddingByte2: PaddingByte
        nwkCoordAddress: t.NWK
        nwkCoordExtAddress: t.EUI64
        nwkPanId: t.uint16_t
        nwkState: NwkState16
        channelList: t.Channels
        beaconOrder: t.uint8_t
        superFrameOrder: t.uint8_t
        scanDuration: t.uint8_t
        battLifeExt: t.uint8_t
        allocatedRouterAddresses: t.uint32_t
        allocatedEndDeviceAddresses: t.uint32_t
        nodeDepth: t.uint8_t
        extendedPANID: t.EUI64
        nwkKeyLoaded: t.Bool
        spare1: t.NwkKeyDesc
        spare2: t.NwkKeyDesc
        spare3: t.uint8_t
        spare4: t.uint8_t
        nwkLinkStatusPeriod: t.uint8_t
        nwkRouterAgeLimit: t.uint8_t
        nwkUseMultiCast: t.Bool
        nwkIsConcentrator: t.Bool
        nwkConcentratorDiscoveryTime: t.uint8_t
        nwkConcentratorRadius: t.uint8_t
        nwkAllFresh: t.uint8_t
        PaddingByte3: PaddingByte  # type:ignore[valid-type]
        nwkManagerAddr: t.NWK
        nwkTotalTransmissions: t.uint16_t
        nwkUpdateId: t.uint8_t
        PaddingByte4: PaddingByte  # type:ignore[valid-type]

    nib = t.NIB(
        SequenceNum=54,
        PassiveAckTimeout=5,
        MaxBroadcastRetries=2,
        MaxChildren=51,
        MaxDepth=15,
        MaxRouters=51,
        dummyNeighborTable=0,
        BroadcastDeliveryTime=30,
        ReportConstantCost=0,
        RouteDiscRetries=0,
        dummyRoutingTable=0,
        SecureAllFrames=1,
        SecurityLevel=5,
        SymLink=1,
        CapabilityFlags=143,
        TransactionPersistenceTime=7,
        nwkProtocolVersion=2,
        RouteDiscoveryTime=13,
        RouteExpiryTime=30,
        nwkDevAddress=0x0000,
        nwkLogicalChannel=25,
        nwkCoordAddress=0x0000,
        nwkCoordExtAddress=t.EUI64.convert("00:00:00:00:00:00:00:00"),
        nwkPanId=0xABCD,
        nwkState=t.NwkState.NWK_ROUTER,
        channelList=t.Channels.CHANNEL_25,
        beaconOrder=15,
        superFrameOrder=15,
        scanDuration=4,
        battLifeExt=0,
        allocatedRouterAddresses=1,
        allocatedEndDeviceAddresses=1,
        nodeDepth=0,
        extendedPANID=t.EUI64.convert("AA:BB:CC:DD:EE:FF:00:11"),
        nwkKeyLoaded=t.Bool(True),
        spare1=t.NwkKeyDesc(KeySeqNum=0, Key=16 * [0]),
        spare2=t.NwkKeyDesc(KeySeqNum=0, Key=16 * [0]),
        spare3=0,
        spare4=0,
        nwkLinkStatusPeriod=15,
        nwkRouterAgeLimit=5,
        nwkUseMultiCast=t.Bool(False),
        nwkIsConcentrator=t.Bool(True),
        nwkConcentratorDiscoveryTime=120,
        nwkConcentratorRadius=10,
        nwkAllFresh=1,
        nwkManagerAddr=0x0000,
        nwkTotalTransmissions=39020,
        nwkUpdateId=0,
    )

    # Make sure all the same fields exist
    assert [f.name for f in t.NIB.fields] == [
        f.name for f in OldNIB.fields if not f.name.startswith("PaddingByte")
    ]

    # Make sure the new NIB can be deserialized by the new NIB struct
    old_nib, remaining = OldNIB.deserialize(nib.serialize(align=True))
    assert not remaining

    # And vice versa
    new_nib, remaining = t.NIB.deserialize(old_nib.serialize(), align=True)
    assert not remaining
    assert new_nib == nib

    # And they are deserialized correctly
    for field in nib.fields:
        assert getattr(nib, field.name) == getattr(old_nib, field.name)


def test_struct_addrmode_address():
    class TestStruct(t.CStruct):
        addr: t.AddrModeAddress

    struct = TestStruct(addr=t.AddrModeAddress(mode=t.AddrMode.NWK, address=0x1234))
    assert struct.get_size(align=False) == 1 + 8
    assert struct.get_size(align=True) == 1 + 8
